// DSPVXD.c

#define   DEVICE_MAIN
#include  "dspvxd.h"
#undef    DEVICE_MAIN

Declare_Virtual_Device(DSPVXD)

// Handle of Synchronization Event between App/VxD
HANDLE		CommonEvent;
// points to ADS56301/HI32 device node in Win95 Reg				
DEVNODE		HI32DeviceNode;
CONFIGRET	RetValue;
// Buffer for HI32's Logical Configuration
CMCONFIG	HI32LogicalConfiguration;	
// Physical Page Add of Base Add of HI32 Memory Space
DWORD		HI32MemSpaceFirstPage;
// Linear Page Add of Base Add of HI32 Memory Space
DWORD		HI32MemSpaceLinAddr;
// Handle for virtual IRQ
IRQHANDLE	HI32_IRQHandle;
// Thunk for interrupt handler
VPICD_HWInt_THUNK HI32_Int_Thunk;
// Thunk for event handler
PriorityVMEvent_THUNK HI32_EventThunk;	

DWORD		Message;
CHAR		ID2[23];
DWORD		HI32MemSpaceLinAdLocked;		
DWORD		LockedLinearAddress;
DWORD*		HCVRAddress;
DWORD*		HSTRAddress;
DWORD		HSTR;
		
///////////////////////////////////////////////////////////////////////////////
VOID SearchHWTree(DEVNODE Node, DEVNODE* TargetNode)
{
        DEVNODE		Child, Sibling;
		ULONG		size;
		PCHAR		ID1;

        if (Node == 0){
                CONFIGMG_Locate_DevNode(&Node, NULL, 0);
                if (Node == 0)
                        return;
        }
		CONFIGMG_Get_Device_ID_Size(&size,Node, 0);
		if (ID1=malloc(size+1))
			CONFIGMG_Get_Device_ID(Node,ID1,size+1,0);		
		if (strncmp(ID1,ID2,strlen(ID2))){
		}
		else{
			*TargetNode = Node;
		}
        if ( CONFIGMG_Get_Child(&Child, Node, 0) != CR_SUCCESS)
                return;
        else {
                SearchHWTree(Child,TargetNode);
                while ( CONFIGMG_Get_Sibling(&Sibling, Child, 0) == CR_SUCCESS ) {
                        SearchHWTree(Sibling,TargetNode);
                        Child = Sibling;
                }
        }
}
//////////////////////////////////////////////////////////////////////////
BOOL	Initial()
{
	// Get device node ID for HI32 device ID
	SearchHWTree((DEVNODE)NULL,&HI32DeviceNode);
	// Get HI32 Logical Configuration Record
	RetValue = CONFIGMG_Get_Alloc_Log_Conf(&HI32LogicalConfiguration,HI32DeviceNode,0);
	if (RetValue == CR_INVALID_DEVNODE)
		return FALSE;
	HI32MemSpaceFirstPage = (DWORD)HI32LogicalConfiguration.dMemBase[0] >> 12;
	// Reserve one page's linear add
	HI32MemSpaceLinAddr = (DWORD) PageReserve(PR_SYSTEM,1,PR_FIXED);
	// Commit reserved linear addresses to physical
	PageCommitPhys(HI32MemSpaceLinAddr >> 12,1, HI32MemSpaceFirstPage,PC_INCR | PC_WRITEABLE | PC_USER);
	// Lock linear pages
	HI32MemSpaceLinAdLocked = LinPageLock(HI32MemSpaceLinAddr >> 12, 1, PAGEMAPGLOBAL);
	return TRUE;
}
//////////////////////////////////////////////////////////////////////////
VOID __stdcall EventService(VMHANDLE hVM, PVOID Refdata, PCLIENT_STRUCT pcrs, DWORD Flags)
{
	if (CommonEvent) {
		_VWIN32_SetWin32Event(CommonEvent);
	}
}

//////////////////////////////////////////////////////////////////////////
// Interrupt Handler
BOOL __stdcall HI32_Int_Handler(VMHANDLE hVM, IRQHANDLE hIRQ)
{
	// check if HINTA bit is set and service the interrupt if it is from the DSP
	HSTRAddress	= (DWORD*)(HI32MemSpaceLinAdLocked) + 0x5;
	if (*HSTRAddress & 0x00000040) {
		// send Host Command for deasserting INTA
		HCVRAddress	= (DWORD*)(HI32MemSpaceLinAdLocked) + 0x6;
//		*HCVRAddress = 0x000000ff;
		*HCVRAddress = 0x000080ff;
		// wait until HINTA bit is reset
		HSTRAddress	= (DWORD*)(HI32MemSpaceLinAdLocked) + 0x5;
		while (HSTR & 0x00000040) {
			HSTR = *HSTRAddress;
		}
//		while (*HSTRAddress & 0x00000040) {}
		// tell VPICD to clear the interrupt
		VPICD_Phys_EOI(HI32_IRQHandle);
		// signal to App that the interrupt has been received and acknowledged
		Call_Priority_VM_Event(TIME_CRITICAL_BOOST,
								Get_Sys_VM_Handle(), 
								0,
								NULL, 
								EventService,
								0,
								&HI32_EventThunk
								);
	}
	return TRUE;
}
//////////////////////////////////////////////////////////////////////////
BOOL	InterruptEnable()
{
	// struct to pass to VPICD_Virtualize_IRQ
	struct VPICD_IRQ_Descriptor IRQdesc;	
	// Fill up the structure to pass to VPICD_Virtualize_IRQ
	// IRQ to virtualize
	IRQdesc.VID_IRQ_Number = (DWORD)HI32LogicalConfiguration.bIRQRegisters[0];
	// Flags
	IRQdesc.VID_Options = 0x17;
	// set address of handler
	IRQdesc.VID_Hw_Int_Proc =		
		(DWORD)VPICD_Thunk_HWInt(HI32_Int_Handler, &HI32_Int_Thunk);
	// The other callbacks are not used.
	IRQdesc.VID_Virt_Int_Proc = 0;
	IRQdesc.VID_EOI_Proc = 0;
	IRQdesc.VID_Mask_Change_Proc = 0;
	IRQdesc.VID_IRET_Proc = 0;
	// Now pass the structure to VPICD. VPICD returns the IRQ handle.
	HI32_IRQHandle = VPICD_Virtualize_IRQ(&IRQdesc);
	// unmask IRQ
	VPICD_Physically_Unmask(HI32_IRQHandle);
	return TRUE;
}
//////////////////////////////////////////////////////////////////////////
DefineControlHandler(SYS_DYNAMIC_DEVICE_INIT, OnSysDynamicDeviceInit);
DefineControlHandler(SYS_DYNAMIC_DEVICE_EXIT, OnSysDynamicDeviceExit);
DefineControlHandler(W32_DEVICEIOCONTROL, OnW32Deviceiocontrol);

BOOL __cdecl ControlDispatcher(
	DWORD dwControlMessage,
	DWORD EBX,
	DWORD EDX,
	DWORD ESI,
	DWORD EDI,
	DWORD ECX)
{
	START_CONTROL_DISPATCH

		ON_SYS_DYNAMIC_DEVICE_INIT(OnSysDynamicDeviceInit);
		ON_SYS_DYNAMIC_DEVICE_EXIT(OnSysDynamicDeviceExit);
		ON_W32_DEVICEIOCONTROL(OnW32Deviceiocontrol);

	END_CONTROL_DISPATCH

	return TRUE;
}


BOOL OnSysDynamicDeviceInit()
{
	CommonEvent = 0;
	return TRUE;
}

BOOL OnSysDynamicDeviceExit()
{
	if (CommonEvent)	{
		_VWIN32_CloseVxDHandle(CommonEvent);
	}
	VPICD_Physically_Mask(HI32_IRQHandle);
	VPICD_Force_Default_Behavior(HI32_IRQHandle);
	LinPageUnLock(HI32MemSpaceLinAdLocked >> 12, 1, PAGEMAPGLOBAL);

	return TRUE;
}

DWORD OnW32Deviceiocontrol(PIOCTLPARAMS p)
{
	DWORD		CfgOffset;
	DWORD*		CfgBuf;
	DWORD		CfgBuf2;
	DWORD		DevID;
	DWORD		VenID;
	CHAR		cDevID[5];
	CHAR		cVenID[5];
	DWORD		LinearAddress;
	DWORD		Size;
	DWORD		PhysicalAddress;
	DWORD		Status;

	switch (p->dioc_IOCtlCode)
	{		
		case DIOC_OPEN:
		case DIOC_CLOSEHANDLE:

			return	0;

		case HI32_USER_MESSAGE:

			Message		=	((DWORD*)(p->dioc_InBuf))[0];
			if (Message) {
				switch (Message) {
					case INITIALIZATION_MESSAGE:	// Initialization Message

						CommonEvent =	((HANDLE*)p->dioc_InBuf)[1];
						LinPageUnLock(HI32MemSpaceLinAdLocked, 1, PAGEMAPGLOBAL);
						VenID			= 0x1057;
						DevID			= ((DWORD*)(p->dioc_InBuf))[2];
						strcpy (ID2,"PCI\\VEN_");
						_ultoa(VenID, cVenID, 16);
						strcat(ID2,cVenID);
						strcat (ID2,"&DEV_");
						_ultoa(DevID, cDevID, 16);
						strcat(ID2,cDevID);
						Status = 0;
						if ( Initial() ) {
							InterruptEnable();
							((DWORD*)(p->dioc_OutBuf))[1] = HI32MemSpaceLinAdLocked;
						}
						else {
							Status = (Status | DEVNODE_NOT_FOUND);
							((DWORD*)(p->dioc_OutBuf))[1] = (DWORD)0; 
						}
						((DWORD*)(p->dioc_OutBuf))[0] = Status;
						*p->dioc_bytesret = 2*sizeof(DWORD);

						return 0;
					
					case LOCK_ONE_MEMORY_PAGE:	// Lock Buffer Pages
						LinearAddress		= ((DWORD*)(p->dioc_InBuf))[1]; 
						Size				= ((DWORD*)(p->dioc_InBuf))[2]; 
						Size = ((LinearAddress + Size - 1) >> 12) - (LinearAddress >> 12) + 1;
						LockedLinearAddress	= LinPageLock(LinearAddress	>> 12,Size, 0);
						if (LockedLinearAddress) {
							((DWORD*)(p->dioc_OutBuf))[0] = (DWORD)Size;	// SUCCEEDED
						}
						else {
							((DWORD*)(p->dioc_OutBuf))[0] = (DWORD)0;	// FAILED
						}
						((DWORD*)(p->dioc_OutBuf))[1] = LockedLinearAddress; 
						*p->dioc_bytesret = 2*sizeof(DWORD);

						return 0;
								
					case RD_CNFG_SPACE_MESSAGE:	// Read ConfigSpace
						CfgOffset		= ((DWORD*)(p->dioc_InBuf))[1];
						CfgBuf = 0;
						RetValue = CONFIGMG_Call_Enumerator_Function(HI32DeviceNode,0,CfgOffset,&CfgBuf,0x4,0);
						if (RetValue == CR_SUCCESS) {
							((BOOL*)(p->dioc_OutBuf))[0] = (BOOL)TRUE;
						}
						else {
							((BOOL*)(p->dioc_OutBuf))[0] = (BOOL)FALSE;
						}
						((DWORD*)(p->dioc_OutBuf))[1] = (DWORD)CfgBuf;
						*p->dioc_bytesret = 2*sizeof(DWORD);

						return 0;

					case WR_CNFG_SPACE_MESSAGE:	// Write ConfigSpace
						CfgOffset		= ((DWORD*)(p->dioc_InBuf))[1];
						CfgBuf2			= (DWORD)((DWORD*)(p->dioc_InBuf))[2];
						RetValue = CONFIGMG_Call_Enumerator_Function(HI32DeviceNode,1,CfgOffset,&CfgBuf2,0x4,0);
						if (RetValue == CR_SUCCESS) {
							((BOOL*)(p->dioc_OutBuf))[0] = (BOOL)TRUE;
						}
						else {
							((BOOL*)(p->dioc_OutBuf))[0] = (BOOL)FALSE;
						}
						((DWORD*)(p->dioc_OutBuf))[1] = (DWORD)*CfgBuf;
						*p->dioc_bytesret = 2*sizeof(DWORD);

						return 0;

					case GET_PHYS_ADD_MESSAGE:	// Convert Linear Address to Physical Address
								// It is supposed that the correspondent page was previously locked (Message 2)
						// Get Linear Address from App
						LinearAddress		= ((DWORD*)(p->dioc_InBuf))[1];
						// Retrieve correspondent Physical Address
						CopyPageTable(LinearAddress	>> 12,1,(PPVOID)&PhysicalAddress,0);
						PhysicalAddress	=	((PhysicalAddress & 0xfffff000) | (LinearAddress & 0x0fff));
						((DWORD*)(p->dioc_OutBuf))[0] = LinearAddress;
						((DWORD*)(p->dioc_OutBuf))[1] = PhysicalAddress;
						*p->dioc_bytesret = 2*sizeof(DWORD);
						return 0;

					case UNLOCK_ONE_MEMORY_PAGE:	// Unlock Buffer Pages (those which were locked in case 2)
						LinearAddress		= ((DWORD*)(p->dioc_InBuf))[1]; // Linear Address from App
						Size				= ((DWORD*)(p->dioc_InBuf))[2]; // Size of Buffer, in bytes
						Size = ((LinearAddress + Size - 1) >> 12) - (LinearAddress >> 12) + 1;
						((BOOL*)(p->dioc_OutBuf))[0] = LinPageUnLock(LinearAddress >> 12, Size, 0);
						*p->dioc_bytesret = sizeof(BOOL);

						return 0;

				}	// switch (Message)

			}	// if (CommonEvent)

			return 0;

		default:
			return	-1;
	}	// switch (p->dioc_IOCtlCode)
	return 0;
}
