#include <stdio.h>
#include <strings.h>

#include "dissas/util56k.h"


/* global hash tables */


#ifdef HASH_BY_NAME
symbol_entry symbol_hash_by_name_table[SYMBOL_TABLE_SIZE];
#endif
symbol_entry symbol_hash_by_addr_table[SYMBOL_TABLE_SIZE][(eNullSpace+1)];

char symbol_names[SYMBOL_TABLE_SIZE][20];
int symbol_name_index;
unsigned int pc;


int read_file(char* app_file,char* input_buffer) {
  FILE *input;
  int size;
  
  if ((input = fopen(app_file, "r")) == NULL) {
    (void) fprintf(stderr, "unable to open file: %s \n", app_file);
    return;
  }
  
  size = fread(input_buffer, 1, 32767, input);  
  fclose(input);
  return size;
}



void put_symbol(char* symbol_name,
		unsigned int mem_address,
		unsigned int mem_space) {
  int hash,i;

  strcpy((char *)&(symbol_names[symbol_name_index][0]), (const char *)symbol_name);
  

#ifdef HASH_BY_NAME
      hash = 0;
      i = 0;
      while (symbol_name[i] != '\0')
	hash += (int)symbol_name[i++];
      hash &= (SYMBOL_TABLE_SIZE-1);
      i = 0;
	while ( symbol_hash_by_name_table[hash].address>=0 && i<SYMBOL_TABLE_SIZE )
	  {
	    hash++;
	    hash &= (SYMBOL_TABLE_SIZE-1);
	    i++;
	  }
	if (i < SYMBOL_TABLE_SIZE)
	  {
	    symbol_hash_by_name_table[hash].name = (char *)&(symbol_names[symbol_name_index][0]);
	    symbol_hash_by_name_table[hash].address = ((mem_space<<16) + mem_address)&0x007FFFFF;
	  }
	else
	  fprintf (stdout, "symbol_hash_by_name_table full. \n");
#endif   /* HASH_BY_NAME */

	
	hash = (mem_address + (mem_address/SYMBOL_TABLE_SIZE))&(SYMBOL_TABLE_SIZE-1);
	i = 0;
	while ( symbol_hash_by_addr_table[hash][mem_space&0x07].address>=0 && i<SYMBOL_TABLE_SIZE ) {
	  hash++;
	  hash &= (SYMBOL_TABLE_SIZE-1);
	  i++;
	}
	if (i < SYMBOL_TABLE_SIZE) {
	  symbol_hash_by_addr_table[hash][mem_space&0x07].name = (char *)&(symbol_names[symbol_name_index][0]);
	  symbol_hash_by_addr_table[hash][mem_space&0x07].address = ((mem_space<<16) + mem_address)&0x007FFFFF;
	}
	else
	  fprintf (stdout, "symbol_hash_by_addr_table full. \n");	
	symbol_name_index++;
	
}


/*  get next line */

char* get_next(char* input_ptr,char* buffer_end) {
    while (input_ptr<buffer_end && *input_ptr!='\n')
      input_ptr++;
    while (input_ptr<buffer_end && isspace(*input_ptr))
      input_ptr++;
    return input_ptr;
}





void hash_symbols(char* input_ptr,char* buffer_end) {
  char symbol_name[20],mode;
  unsigned int mem_address,mem_space;
  int i;
  
  while (input_ptr<buffer_end && isspace(*input_ptr)) 
    input_ptr++;

  while (input_ptr < buffer_end) {
    if (*input_ptr == '_') { 
      if (strncmp(input_ptr, "_SYMBOL ", 8) == 0) {    
	
	/* determine the memory space */
	
	input_ptr += 8;
	mode = *input_ptr++;
	
	if (mode == 'P') {
	    mem_space = ePSpace; 
	    mode = SYMBOL;
	}
	else if (mode == 'X') {
	  mem_space = eXSpace; 
	  mode = SYMBOL;
	}
	else if (mode == 'Y') {
	  mem_space = eYSpace; 
	  mode = SYMBOL;
	}
	else if (mode == 'L') {
	  mem_space = eLSpace; 
	  mode = SYMBOL;
	}
	else if (mode == 'N') {
	  mem_space = eNullSpace; 
	  mode = SYMBOL;
	}
	else
	  mode = OTHER;
	
	/* read the symbols  */

	input_ptr = get_next(input_ptr,buffer_end);
      }
    }
  

    if (mode == SYMBOL) {
      i = 0;
      while (!isspace(*input_ptr) && i<19)
	symbol_name[i++] = *(input_ptr++);
      symbol_name[i] = '\0';
      while (strncmp(input_ptr, " I ", 3) != 0)
	input_ptr++;
      input_ptr += 3;
      sscanf(input_ptr, "%6x", &mem_address);
      mem_address &= 0x0000FFFF;
    
      /* hash the symbol */
      
      put_symbol(symbol_name,mem_address,mem_space);
      
      
      
    }

    input_ptr = get_next(input_ptr,buffer_end);

  }
}





int dissas(char mode,char* input_ptr,char* buffer_end) {
  char op_str[6];
  int  op,op2;
  char* nextcom;
  char output_string[256];
  char symbol_name[256];
  int i,num_ops = 1;


  switch (mode)
    {
    case SYMBOL:
      break;      
    case 'X':
    case 'Y':
    case 'P':
      for (i=0; i<6; i++)
	op_str[i] = *input_ptr++;
      
      sscanf(op_str, "%6x", &op);
      op&=0x00FFFFFF;

      nextcom = get_next(input_ptr,buffer_end);

      for (i=0; i<6; i++)
	op_str[i] = *nextcom++;
      
      sscanf(op_str, "%6x", &op2);
      op2&=0x00FFFFFF;

      /* dissasemble */

      num_ops = Disassemble(op, op2, output_string);  

      i = FindSymbol(ePSpace, pc, symbol_name);					
       
      if (i >= 0)
	  fprintf(stdout,"%s\n", symbol_name);
      
      if (num_ops > 1) {
     	fprintf(stdout, " p:%04X  %06X %06X    %s\n", pc, op, op2, output_string);
	pc++;
      }
      else
	fprintf(stdout, " p:%04X  %06X           %s\n", pc, op, output_string);

      pc++;

      break;
      
    case OTHER:
    default:
      while (input_ptr<buffer_end && (*input_ptr!='_' || *(input_ptr-1)!='\n'))
	input_ptr++;
      break;
    }
  
  while (input_ptr<buffer_end && isspace(*input_ptr))
    input_ptr++;

  return num_ops;

}







main(int argc, char** argv) {
  
  int op, input_size,nops;
  unsigned int mem_space = 0, mem_address;
  register int i;
  register char *input_ptr;
  char *buffer_end, op_str[8];
  char input_buffer[32767];
  char mode;
  int j;  

  op_str[6] = '\0';

  /* Initialize hash tables */
  
#ifdef HASH_BY_NAME
  for (i=0; i<SYMBOL_TABLE_SIZE; i++)
    {
      symbol_hash_by_name_table[i].name = NULL;
      symbol_hash_by_name_table[i].address = -1L;
    }
#endif
  
  for (mem_space=eXSpace; mem_space<=eNullSpace; mem_space++)
    
    for (i=0; i<SYMBOL_TABLE_SIZE; i++)  {
      symbol_hash_by_addr_table[i][mem_space].name = NULL;
      symbol_hash_by_addr_table[i][mem_space].address = -1L;
    }
  
  

  /* Read the file */

  for (i=1;i<argc;i++) {
    
    printf("Disassemble of file %s \n\n",argv[i]);
    input_size = read_file(argv[i],input_buffer);
    input_ptr = input_buffer;
    buffer_end = input_buffer + (int)input_size;
    
    hash_symbols(input_ptr,buffer_end);
    
    
    /* start disassembling */
    
    
    mode = OTHER;
    
    while (input_ptr<buffer_end && isspace(*input_ptr))
      input_ptr++;
    
    while (input_ptr < buffer_end) {
      
      /* determine the mode */
      
      if (*input_ptr == '_') {
	if (strncmp(input_ptr, "_DATA ", 6) == 0) {
	  input_ptr += 6;
	  mode = *input_ptr++;
	  if (mode == 'P' || mode == 'X' || mode == 'Y') {
	    while (input_ptr<buffer_end && isspace(*input_ptr))
	      input_ptr++;
	    sscanf(input_ptr, "%4x", &pc);
	  }
	  else {
	    fprintf(stderr, "ERROR - unknown _DATA block type: %c\n", mode);
	    mode = OTHER;
	  }
	  fprintf(stdout,"      org %c: $%06x\n",mode,pc);
	}
	else
	  mode = OTHER;
	
	input_ptr = get_next(input_ptr,buffer_end);
	continue;
      }
      
      
      /* dissassemble  */
      
      nops = 1;
      nops = dissas(mode,input_ptr,buffer_end);
      
      input_ptr = get_next(input_ptr,buffer_end);
      if (nops > 1) 
	input_ptr = get_next(input_ptr,buffer_end);

    }
  }  
}











