#include "multisound.h"
#include "dissas/util56k.h"


#ifdef __linux__
#include<asm/io.h>
#define inp(x) inb(x)
#include "linkbd.h"
#endif



#ifdef HASH_BY_NAME
extern symbol_entry symbol_hash_by_name_table[];
#endif
extern symbol_entry symbol_hash_by_addr_table[][(eNullSpace+1)];

extern char symbol_names[][20];
int float_regs = 1;

#define NUMREG 42
#define NUMBRK 10
#define USTK 2*(NUMREG+NUMBRK)

#define _reg(n) (reg[n]==reg[NUMREG+n]?' ':'*'),reg[n]

#define R0 _reg(0)
#define R1 _reg(1)
#define R2 _reg(2)
#define R3 _reg(3)
#define R4 _reg(4)
#define R5 _reg(5)
#define R6 _reg(6)
#define R7 _reg(7)
#define N0 _reg(8)
#define N1 _reg(9)
#define N2 _reg(10)
#define N3 _reg(11)
#define N4 _reg(12)
#define N5 _reg(13)
#define N6 _reg(14)
#define N7 _reg(15)
#define M0 _reg(16)
#define M1 _reg(17)
#define M2 _reg(18)
#define M3 _reg(19)
#define M4 _reg(20)
#define M5 _reg(21)
#define M6 _reg(22)
#define M7 _reg(23)
#define X0 _reg(24)
#define X1 _reg(25)
#define Y0 _reg(26)
#define Y1 _reg(27)
#define A2 _reg(28)&0xFF
#define A1 _reg(29)
#define A0 _reg(30)
#define B2 _reg(31)&0xFF
#define B1 _reg(32)
#define B0 _reg(33)
#define PC _reg(34)
#define SR _reg(35)
#define OMR _reg(36)
#define LA _reg(37)
#define LC _reg(38)
#define SP _reg(39)
#define BCR _reg(40)
#define IPR _reg(41)




/*
           X1= 000000 X0= 000000  R7= 0000  N7= 0000  M7= FFFF
           Y1= 000000 Y0= 000000  R6= 0000  N6= 0000  M6= FFFF
 A2= 00    A1= 000000 A0= 000000  R5= 0000  N5= 0000  M5= FFFF
 B2= 00    B1= 000000 B0= 000000  R4= 0000  N4= 0000  M4= FFFF
                                  R3= 0000  N3= 0000  M3= FFFF
 PC=*0040  SR= 0300  OMR= 02      R2= 0000  N2= 0000  M2= FFFF
 LA= 0000  LC= 0000   SP= 00      R1= 0000  N1= 0000  M1= FFFF
          BCR= 0000  IPR= 000000  R0= 0000  N0= 0000  M0= FFFF
*/


int Disassemble(unsigned int, unsigned int, char *);

double bit24_to_float(int m);


double bit24_to_float(int m)
{
  double t = (double)m/8388608.0;
  if (m >= 0x800000)
    t -= 2.0;

  return t;
}


double bit48_to_float(int* r)
{
  long long temp;
  double t;
  /*  temp = ((long long)(r[0]&0xff)<<48) + ((long long)r[1]<<24) + r[2];  */
  temp = ((long long)r[1]<<24) + r[2];
  t = (double) temp/(double) (1ll<<47l);
  if (temp >= 0x800000000000ll)
    t-= 2.0;
  return t;
}


void display_reg(int *reg, int REGBASE)
{
	register short i;
	int data, op, op2, pc;
	char output_string[64];
	int num_ops, n;
	static char previous_instruction[80] = {'\0'};
	char symbol_name[20];
	double X0f,Y0f,X1f,Y1f,Af,Bf;

	

	(void) set_dsp_read_ptr(P, REGBASE);

	for (i=0; i<0x89; i++)
		{
		(void) read_dsp_data(&data);
		reg[i] = data&0x00FFFFFFL;
/*
(void) fprintf(stderr, "reg[%02lX] = %06lX \n", i, reg[i]);
*/
		}

	X0f = bit24_to_float(reg[24]);
	Y0f = bit24_to_float(reg[26]);
	X1f = bit24_to_float(reg[25]);
	Y1f = bit24_to_float(reg[27]);
	Af = bit48_to_float(reg+28);
	Bf = bit48_to_float(reg+31);


	if (float_regs) {
	fprintf(stderr," x1= %1.8f x0= %1.8f    r7=%c%04X  n7=%c%04X  m7=%c%04X\n",X1f,X0f, R7, N7, M7);
	fprintf(stderr," y1= %1.8f y0= %1.8f    r6=%c%04X  n6=%c%04X  m6=%c%04X\n",Y1f,Y0f, R6, N6, M6);

	fprintf(stderr," a = %1.9f                  r5=%c%04X  n5=%c%04X  m5=%c%04X\n",Af, R5, N5, M5 );
	fprintf(stderr," b = %1.9f                  r4=%c%04X  n4=%c%04X  m4=%c%04X\n",Bf, R4, N4, M4 );

	}
	else
	  {
	    (void) fprintf(stderr, "           x1=%c%06X x0=%c%06X  r7=%c%04X  n7=%c%04X  m7=%c%04X\n", X1, X0, R7, N7, M7);
	    (void) fprintf(stderr, "           y1=%c%06X y0=%c%06X  r6=%c%04X  n6=%c%04X  m6=%c%04X\n", Y1, Y0, R6, N6, M6);
	    (void) fprintf(stderr, " a2=%c%02X    a1=%c%06X a0=%c%06X  r5=%c%04X  n5=%c%04X  m5=%c%04X\n", A2, A1, A0, R5, N5, M5);
	    (void) fprintf(stderr, " b2=%c%02X    b1=%c%06X b0=%c%06X  r4=%c%04X  n4=%c%04X  m4=%c%04X\n", B2, B1, B0, R4, N4, M4);
	  }


	(void) fprintf(stderr, "                                  r3=%c%04X  n3=%c%04X  m3=%c%04X\n", R3, N3, M3);
	(void) fprintf(stderr, " pc=%c%04X  sr=%c%04X  omr=%c%02X      r2=%c%04X  n2=%c%04X  m2=%c%04X\n", PC, SR, OMR, R2, N2, M2);
	(void) fprintf(stderr, " la=%c%04X  lc=%c%04X   sp=%c%02X      r1=%c%04X  n1=%c%04X  m1=%c%04X\n", LA, LC, SP, R1, N1, M1);
	(void) fprintf(stderr, "          bcr=%c%04X  ipr=%c%06X  r0=%c%04X  n0=%c%04X  m0=%c%04X\n", BCR, IPR, R0, N0, M0);



	(void) fprintf(stderr, "\n");
	
	if (*previous_instruction != '\0')
		(void) fprintf(stderr, " %s", previous_instruction);
	
	
	pc = reg[34];

	for (n=0; n<7; n++)
	  {	
	    (void) set_dsp_read_ptr(P, pc);
		(void) read_dsp_data(&op);
		(void) read_dsp_data(&op2);
		
		op  &= 0x00FFFFFF;
		op2 &= 0x00FFFFFF;
		
		num_ops = Disassemble(op, op2, output_string);
		
		i = FindSymbol(ePSpace, pc, symbol_name);
		if (!n) fprintf(stderr,">");
		else fprintf(stderr," ");
		if (i >= 0)
		  if (num_ops > 1)
		    {
		      fprintf(stderr, " p:%04X  %06X %06X %10s %s\n", pc, op, op2,symbol_name, output_string);
		      if (!n) 
			sprintf(previous_instruction, " p:%04X  %06X %06X %10s %s\n", pc, op, op2,symbol_name, output_string);

		      pc += 2;
			}
		  else
			{
			fprintf(stderr, " p:%04X  %06X        %10s %s\n", pc, op,symbol_name, output_string);
			if (!n) 
			  sprintf(previous_instruction, " p:%04X  %06X        %10s %s\n", pc, op,symbol_name, output_string);

			  
			pc++;
			}
		else
		  if (num_ops > 1)
		    {
		      fprintf(stderr, " p:%04X  %06X %06X            %s\n", pc, op, op2, output_string);
		      if (!n)
			sprintf(previous_instruction, " p:%04X  %06X %06X            %s\n", pc, op, op2, output_string);
		      pc += 2;
		    }
		  else
		    {
		      fprintf(stderr, " p:%04X  %06X                   %s\n", pc, op, output_string);
		      if (!n)
		      sprintf(previous_instruction, " p:%04X  %06X                   %s\n", pc, op, output_string);

		      pc++;
		    }
		
	  }
	
	(void) fprintf(stderr, "\n");
	
	}






void trace(int a) {
  int ISR;
  int i;

  ISR = inp(isrRegister);
  if (ISR&HF2)
    {
      (void) host_command(TRACE_CMD);
      
      for (i=0; i<100000L; i++)
	;
      
    }
  
}


void break_command(int d,int reg) {
  int ISR;
  int i;
  
  ISR = inp(isrRegister);
  if (ISR&HF2)
    {
      
      (void) set_dsp_write_ptr(P, (reg + 0x0054L));
      (void) write_dsp_data(d);
      
      (void) host_command(NMI_CMD);
      
      for (i=0; i<100000L; i++)
	;
    }
}





void dspbugger(char cmd)
	{
	int i, reg[0x89], n,timeout;
	char input_buffer[INPUT_BUFFER_SIZE], char0, char1;
	char label[32];
	static int REGBASE=0;

	int data, mem_address = 0;
	char mem_space = 'p', a_char;
	

	char symbol_name[20];
	int op, op2;
	char output_string[64];
	int pc;
	int num_ops;


	int ISR, oldISR;

	const char reg_name[] = "r0r1r2r3r4r5r6r7n0n1n2n3n4n5n6n7m0m1m2m3m4m5m6m7x0x1y0y1a2a1a0b2b1b0pcsromlalcspbcip";

/*	include this code to read and write to various DSP locations */

	
	if (REGBASE == 0)
		{
		#define STACK_ERROR 0x0002
		set_dsp_read_ptr(P, (STACK_ERROR+1));
		read_dsp_data(&REGBASE);
		REGBASE += 0x0002;
		}

	if ( (cmd == 'w') || (cmd == 'x')  || (cmd == 'c')  || (cmd == 'b') )
		cmd = ' ';

	ISR = inp(isrRegister);
	if ( (ISR&HF2) && (cmd != 'g') && (cmd != 't') )
		{
		  fprintf(stderr, "\n   DSP halted ... Press g to start execution\n");
		  break_command(0x40,REGBASE);
		  display_reg(reg, REGBASE); 

		/*		cmd = 'd'; */
		}

	if (cmd != ' ')
		{
		do
			{
			oldISR = ISR;
			switch (cmd)
				{
				case 'f':
				  float_regs = !float_regs;
				  display_reg(reg, REGBASE);
				  break;
				case 'w':
				  sscanf(input_buffer, "w %c %x", &mem_space, &mem_address);
				  fprintf(stderr,"writing at memory %x\n",mem_address);
				  if (mem_space == 'p')
				    (void) set_dsp_write_ptr(P, mem_address);
				  else if (mem_space == 'y')
				    (void) set_dsp_write_ptr(Y, mem_address);
				  else if (mem_space == 'x')
				    (void) set_dsp_write_ptr(X, mem_address);
				  else
				    break;
			
				(void) fprintf(stderr, "enter data: $");
				(void) fflush(stderr);
				(void) fgets(input_buffer,INPUT_BUFFER_SIZE,stdin);
				(void) sscanf(input_buffer, "%x", &data);
	
				(void) write_dsp_data(data);

				break;
	
				case 'p':
				  if (sscanf(input_buffer,"p %s",label)) {
				    if (hash_symbol(label) != -1) {
				      get_at_label(label,0);
				      (void) read_dsp_data(&data);
				      (void) fprintf(stderr, " %06lX\n", data&0x00FFFFFFL);
				    }
				  }
				  break;
				case 's':
				  sscanf(input_buffer, "s %s", label);
				  if (hash_symbol(label) != -1) {
				    fprintf(stderr,"writing at label %s\n",label);
				    put_at_label(label,0);

				    (void) fprintf(stderr, "enter data: $");
				    (void) fflush(stderr);
				    (void) fgets(input_buffer,INPUT_BUFFER_SIZE,stdin);
				    (void) sscanf(input_buffer, "%x", &data);
				  
				    (void) write_dsp_data(data);
				  }

				break;

				case 'x':
				(void) sscanf(input_buffer, "r %c %x", &mem_space, &mem_address);
	
				if (mem_space == 'p')
					(void) set_dsp_read_ptr(P, mem_address);
				  else if (mem_space == 'y')
					(void) set_dsp_read_ptr(Y, mem_address);
				  else if (mem_space == 'x')
					(void) set_dsp_read_ptr(X, mem_address);
				  else
					break;

				for (i=0; i<8; i++)
					{
					(void) read_dsp_data(&data);
					(void) fprintf(stderr, " %06lX", data&0x00FFFFFFL);
					}
				(void) fprintf(stderr, "\n");
				for (i=0; i<8; i++)
					{
					(void) read_dsp_data(&data);
					(void) fprintf(stderr, " %06lX", data&0x00FFFFFFL);
					}
				(void) fprintf(stderr, "\n");
				break;
	
				case 'd':
				  ISR = inp(isrRegister);
				  if (!(ISR&HF2)) 
				    (void) host_command(NMI_CMD);
				  
				  /* GGeiger */

				  timeout = TIMEOUT;
				  while (!(ISR&HF2) && (timeout > 0)) {
				    ISR = inp(isrRegister);
				    timeout--;;
				  }
				  if (timeout <= 0)
				    fprintf(stderr,"TIMEOUT WAITING FOR DEBUG MODE");
				  
				  fprintf(stderr,"display registers\n");
				  (void) display_reg(reg, REGBASE);
				  
				  oldISR = ISR;
				  break;
				  
				case 'l':
				(void) sscanf(input_buffer, "l %x", &pc);
				for (n=0; n<16; n++)
					{	
					(void) set_dsp_read_ptr(P, pc);
					(void) read_dsp_data(&op);
					(void) read_dsp_data(&op2);
					op  &= 0x00FFFFFF;
					op2 &= 0x00FFFFFF;
					num_ops = Disassemble(op, op2, output_string);
					
					i = FindSymbol(ePSpace, pc, symbol_name);					
					if (i >= 0)
						{
						fprintf(stderr,        "                  %s\n", symbol_name);
						}
					
					if (num_ops > 1)
						{
						(void) fprintf(stderr, " p:%04X  %06X %06X    %s\n", pc, op, op2, output_string);
						pc += 2;
						}
					  else
						{
						(void) fprintf(stderr, " p:%04X  %06X           %s\n", pc, op, output_string);
						pc++;
						}
					}
				break;

				case 't':
				  trace(1);
				  display_reg(reg, REGBASE);

				break;

				case 'g':
				ISR = inp(isrRegister);
				if (ISR&HF2)
					{
					(void) set_dsp_write_ptr(P, (REGBASE + 0x0054L));
					(void) write_dsp_data(0);

					(void) host_command(NMI_CMD);

					for (i=0; i<100000L; i++)
						;
					oldISR = 0;
					}
				break;


				case 'c':
				ISR = inp(isrRegister);
				if (ISR&HF2)
					{
					input_buffer[4] = ' ';
					(void) sscanf(input_buffer, "c %c%c %x", &char0, &char1, &data);

					n = -1;
					for (i=0; i<NUMREG; i++)
						{
						if ( ((int)char0 == reg_name[2*(int)i]) && ((int)char1 == reg_name[2*(int)i+1]) )
							n = i;
						}

					if (n > 0)
						{
						reg[(int)n] = data;
						(void) set_dsp_write_ptr(P, (REGBASE + n));
						(void) write_dsp_data(data);
						}
					  else
						(void) fprintf(stderr, "register name invalid\n");
					}
				break;

				case 'u':
				ISR = inp(isrRegister);
				if (ISR&HF2)
					{
					if (reg[39] == 0)
						(void) fprintf(stderr, "empty stack\n");
					  else
						{
						for (i=0; i<15; i++)
							{
							if (15-i == reg[39])
								(void) fprintf(stderr, "sp-> %04X  %04X\n", reg[2*(int)i+USTK], reg[2*(int)i+USTK+1]);
							  else if (15-i < reg[39])
								(void) fprintf(stderr, "     %04X  %04X\n", reg[2*(int)i+USTK], reg[2*(int)i+USTK+1]);
							}
						}
					}
				break;

				case 'b':
				  (void) sscanf(input_buffer, "b %x", &data);
				  break_command(data,REGBASE);
				  oldISR = 0;
				
				break;

				case '?':
				(void) fprintf(stderr, "  >f               toggle float registers\n");
				(void) fprintf(stderr, "  >x x|y|p <addr>  examine DSP memory\n");
				(void) fprintf(stderr, "  >w x|y|p <addr>  write DSP memory\n");
				(void) fprintf(stderr, "  >d               enter debug and display user registers\n");
				(void) fprintf(stderr, "  >l <addr>        list DSP code\n");
				(void) fprintf(stderr, "  >c <reg> <val>   change user registers\n");
				(void) fprintf(stderr, "  >p <symbol>      print symbol value\n");
				(void) fprintf(stderr, "  >s <symbol>      set symbol value\n");
				(void) fprintf(stderr, "  >u               display user stack\n");
				(void) fprintf(stderr, "  >t               trace 1 instruction (resume)\n");
				(void) fprintf(stderr, "  >b <addr>        set breakpoint and go (resume)\n");
				(void) fprintf(stderr, "  >g               go (resume)\n");
				(void) fprintf(stderr, "  >q               exit debugger\n");
				(void) fprintf(stderr, "  >?               display debugger commands\n");
				break;
				}

			(void) fprintf(stderr, ">");
			(void) fflush(stderr);


			if ( !(oldISR&HF2) )
				{
				while (!kbhit())
					{
					ISR = inp(isrRegister);
					if ( (ISR&HF2) )
						{
						(void) fprintf(stderr, "  **DSP halted**\n");
						(void) display_reg(reg, REGBASE);
						(void) fprintf(stderr, ">");
						(void) fflush(stderr);
						goto out;
						}
					}
				}

		out:
			a_char = '\n';
			(void) fgets(input_buffer,INPUT_BUFFER_SIZE,stdin);
			(void) sscanf(input_buffer, "%c", &a_char);
			if (a_char != '\n')
				cmd = a_char;

			} while (cmd != 'q');
		}
	}


















































