;
;
;	multisou.asm
;
;	DSP56K code to facilitate SSI sample I/O in CHUNK_SIZE sample chunks.
;
;	Includes an image of host_port.asm (to facilitate host <-> DSP
;	communications) in low internal boot memory that is moved to
;	high external memory by the boot code.
;
;	Copyright  1993 Robert Bristow-Johnson
;	Copyright  1994, 1995 crescent engineering
;
;

		include	'56k.inc'


CHUNK_SIZE	equ	4
sys_code	equ	p:$7c00
;
;
;
;		000000		nop
;		0BF080		jsr	>$babe
;		00BABE
;		085FAB		movep		x:<<HRX,x:(r7)+
;		085FEB		movep		x:<<HRX,y:(r7)+
;		085F6B		movep		x:<<HRX,p:(r7)+
;		08DFAB		movep		x:(r7)+,x:<<HTX
;		08DFEB		movep		y:(r7)+,x:<<HTX
;		08DF6B		movep		p:(r7)+,x:<<HTX
;
;
;		Equates for 56k opcodes
;
NOP_OP		equ	$000000
JSR_OP		equ	$0BF080
RX_x_OP		equ	$085FAB
RX_y_OP		equ	$085FEB
RX_p_OP		equ	$085F6B
TX_x_OP		equ	$08DFAB
TX_y_OP		equ	$08DFEB
TX_p_OP		equ	$08DF6B




		org	x:$3F
host_save		ds	1

		org	y:$3F
buffer_ptr	ds	1

		org	x:$40
left_input	ds	CHUNK_SIZE
right_input	ds	CHUNK_SIZE

;		org	x:$40			;outputs can be same or different than inputs
left_output	ds	CHUNK_SIZE
right_output	ds	CHUNK_SIZE

		org	y:$40
SSI_buffer	ds	(4*CHUNK_SIZE)


		org	p:P_RESET
	jmp 	>boot				;RESET

		org	p:HOST_RX
	nop					;HOST RX
	nop
	nop					;HOST TX
	nop
	jsr	>set_dsp_read_write_ptr	        ;HOST CMD
	jsr	>set_input_pot_gain
	jsr	>set_aux_pot_gain
	
	
		org	p:SSI_RX
	movep			y:(r0),x:<<TX	;SSI_RX
	movep			x:<<RX,y:(r0)+
	movep			y:(r0),x:<<TX	;SSI_RX_ERROR
	movep			x:<<RX,y:(r0)+


		org	p:$3F

process_stereo_chunk
	
	nop

;		org	p:$40
;	
;	PUT YOUR SAMPLE PROCESSING CODE HERE!
;
;	The sample processing code can expect to have CHUNK_SIZE valid stereo
;	samples located at x:left_input and x:right_input.  It must process them
;	in whatever way it's designed to and leave the resulting
;	samples in x:left_output and x:right_output.
;
;	The scaling mode going in is NO_SCALING.  The interrupt mask
;	is at its lowest setting (ENABLE_INTS) enabling all interrupts
;	including host command, receive, and transmit.  If it is desired
;	to hold off host interrupts during the chunk processing time, the
;	interrupt mask may be bumped up to DISABLE_IPL_0 or DISABLE_IPL_1
;	but no higher (i.e. DISABLE_IPL_2) because SSI interrupts, that
;	are at IPL 2, must be allowed to be serviced.
;
;	Registers r7, n7, and m7 are dedicated for host port communications
;	and are not available for use by the user (except m7, which is
;	always linear ($FFFF), can be copied to other m registers to save
;	a clock pulse compared to an immediate move).  
;
;	Registers r0, n0, and m0 are used by the SSI interrupt handler and
;	are unavailable for use by the user.
;
;	Registers a, x0, r1, r2, r6 (and m1, m2, m6) are changed upon entry.
;	All other registers are unchanged from exiting the last chunk.
;
;	After all CHUNK_SIZE stereo output samples have been computed, the
;	sample chunk processing code can terminate using the rti instruction
;	(instead of rts).  This returns the scaling mode and interrupt
;	level back to their previous settings and returns to the main_loop.
;
	
	movec			m7,m1				;make sure modulo's are linear
	movec			m7,m2
	
	move			#<left_input,r1
	move			#<left_output,r2
	do	#CHUNK_SIZE,left_loop
	move			x:(r1)+,x0
	move			x0,x:(r2)+
left_loop

	move			#<right_input,r1
	move			#<right_output,r2
	do	#CHUNK_SIZE,right_loop
	move			x:(r1)+,x0
	move			x0,x:(r2)+
right_loop
	
	rti							;null sample chunk processing



;
;	This startup code can be written over with sample processing code
;	as it is only used upon reset and the host is disabled to write
;	to anywhere until the sys_code image is moved and interrupts
;	are enabled.
;

boot
	movec	#$0300,sr	;Initialize SR to block interrupts

	movec	#0,sp		;Set stack pointer to zero
	movec	#DE,omr		;0
				;0	STOP delay = 65536 clock cycles
				;000	unused
				;1	Enable u-law and sin-wave rom tables
				;00	Operating Mode = 0


	movep	#$1111,x:BCR	; setup timing for external bus


;---    sci and ssi initialization      ---
	movep	#0,x:PCC		; zero PCC to cycle it
	movep	#$0302,x:SCR	; sci register
	movep	#$0013,x:SCCR	; clk divider for MIDI baud rate
	movep	#$0004,x:PCDDR	; init port c data direction
				; set word len=24 bits, 2 words/frame
	movep	#$6100,x:CRA	; init ssi control a
	movep	#$B800,x:CRB	; init ssi control b
	movep	#0,x:TX		; zip the tx ssi reg
	movep	#$FFFFFB,x:PCC	; init port c control register

;---    Init the interrupts             ---

	movep	#$7400,x:IPR	; init the interrupt priority register

;---    Init the external perphs       ---

	move	#>$0000,y0	; get 44k mask
	move	y0,y:$FFC2	; set counter
	move	y0,y:$FFC3	; set counter

	movep	#>$000011,y:$FFC1	; reset dad
	movep	#>$000000,y:$FFC1	; let run dad
	bset	#0,y:$FFC4	; allow for writes in d0-d7
	
	movep	#>$0080,y:$FFC0	; enable digital pot




	movep	#$0007,x:HCR	;Enable interrupt for host command, rx, tx
				;Clear HF3 to signal to host not to read
	movep	#$0001,x:PBC	;Enable host port


	move	#<SSI_buffer,r0
	move	#<(2*CHUNK_SIZE),n0
	movec	#<(4*CHUNK_SIZE-1),m0
	clr	a		r0,y:<buffer_ptr


	movep			a,x:<<HTX		;fill host -> DSP buffer
	movep			a,x:<<HTX

	movec	#$FFFF,m7
	movec	m7,m1
	movec	m7,m2
	movec	m7,m3
	movec	m7,m4
	movec	m7,m5
	movec	m7,m6


	clr	a		#<0,r1
	rep	#$100
	move		a,l:(r1)+

	move			#>$8000,n1
	move			#>$4000,r1
	rep	n1
	move		a,x:(r1)+



	move			#<sys_code_image,r1
	move			#sys_code,r2

	do	#(sys_code_image_end-main_loop),move_sys_code_loop
;	do      #10,move_sys_code_loop
	movem	p:(r1)+,x1
	movem	x1,p:(r2)+
move_sys_code_loop

	
	move	#>$292900,a1		; set initial gain value
	move	#>$0082,x0		; set "amps on" bit, select IN pot
	jsr	set_digital_pot	; set the IN pot gain

	move	#>$6B6B00,a1		; set initial gain value
	move	#>$0084,x0		; set "amps on" bit, select AUX pot
	jsr	set_digital_pot	; set the AUX pot gain
	

	move	#<0,r1			; clear all user registers
	move	#<0,r2
	move	#<0,r3
	move	#<0,r4
	move	#<0,r5
	move	#<0,r6
	move	#<0,r7

	move	#<0,n1
	move	#<0,n2
	move	#<0,n3
	move	#<0,n4
	move	#<0,n5
	move	#<0,n6
	move	#<0,n7

	clr	a	#<0.0,x0
	move		#<0.0,x1
	clr	b	#<0.0,y0
	move		#<0.0,y1

	andi	#ENABLE_INTS,mr		;bump down interrupt mask

	jmp	main_loop




sys_code_image
;
;
;	This image of the system code below is appended to
;	the boot code in low internal p: memory and not executed
;	until it is moved to the correct location at p:sys_code.
;

	org	p:sys_code

main_loop
	movec		m7,m1			;make sure modulo's are linear
	movec		m7,m2
	movec		m7,m6
	move			y:<buffer_ptr,r6	;get guzzouta location in buffer
	move		#<left_output,r1
	move		#<right_output,r2
	move		x:(r1)+,a			;get left
	move		x:(r2)+,a	a,y:(r6)+		;get right, store zeroeth outgoing sample

;	dup	CHUNK_SIZE-1
;	move		x:(r1)+,a	a,y:(r6)+		;get left, store first outgoing sample
;	move		x:(r2)+,a	a,y:(r6)+		;... , store second outgoing sample
;	endm
; replaced this macro by

	move		x:(r1)+,a	a,y:(r6)+		;get left, store first outgoing sample
	move		x:(r2)+,a	a,y:(r6)+		;... , store second outgoing sample
	move		x:(r1)+,a	a,y:(r6)+		;get left, store first outgoing sample
	move		x:(r2)+,a	a,y:(r6)+		;... , store second outgoing sample
	move		x:(r1)+,a	a,y:(r6)+		;get left, store first outgoing sample
	move		x:(r2)+,a	a,y:(r6)+		;... , store second outgoing sample



	move			a,y:(r6)+

	move		#>SSI_buffer,x1
wait_for_chunk_start				;make sure we get past
	move		r0,a			;  the zeroeth sample of
wait_for_chunk_start2
	sub	x1,a	n0,x0
	jeq	wait_for_chunk_start		;  looking for the next chunk
	cmp	x0,a	r0,a			;  the current chunk before
	jeq	wait_for_chunk_start2		;  looking for the next chunk
wait_for_next_chunk
	sub	x1,a				;test for zeroeth sample
	jeq	begin_next_chunk
	cmp	x0,a	r0,a			;test for zeroeth sample
	jne	wait_for_next_chunk

begin_next_chunk
	jclr	#RFS,x:SSISR,rxR_txL		;make sure we stay in
wait_for_next_sample				;  sync to rx left/tx right
	jclr	#0,r0,wait_for_next_sample
	move		(r0)-			;back up r0 in case we did have to sync


rxR_txL						;we are now at the zeroeth sample
						;  the next chunk
	lua	(r0)-n0,r6			;go back CHUNK_SIZE samples
	move			r6,y:<buffer_ptr	;today's guzzinta is tommorow's guzzouta
	move		#<left_input,r1
	move		#<right_input,r2
	move			y:(r6)+,a		;get zeroeth incoming sample
	move		a,x:(r1)+	y:(r6)+,a		;store left, get first sample

	
;	dup
;	move		a,x:(r2)+	y:(r6)+,a		;store right, get second sample
;	move		a,x:(r1)+	y:(r6)+,a		;store left, etc.
;	endm
; Replaced this macro by

	move		a,x:(r2)+	y:(r6)+,a		;store right, get second sample
	move		a,x:(r1)+	y:(r6)+,a		;store left, etc.
	move		a,x:(r2)+	y:(r6)+,a		;store right, get second sample
	move		a,x:(r1)+	y:(r6)+,a		;store left, etc.
	move		a,x:(r2)+	y:(r6)+,a		;store right, get second sample
	move		a,x:(r1)+	y:(r6)+,a		;store left, etc.
;
;
;
	move		a,x:(r2)+



	jsr	<process_stereo_chunk	;go process left and right

	jmp	main_loop		;loop back indefinitely



;******************************************************************************
;
;	set_dsp_read_write_ptr(long mode_plus_ptr, unsigned char * base_addr);
;
;		mode = 0x00xxxx	read X
;		mode = 0x01xxxx	read Y
;		mode = 0x02xxxx	read P
;		mode = 0x80xxxx	write X
;		mode = 0x81xxxx	write Y
;		mode = 0x82xxxx	write P
;
;		xxxx = pointer address
;
;	host command SET_PTR_HOST_CMD (0x12 + 0x80) is at vector p:$0024
;
;******************************************************************************

set_dsp_read_write_ptr
	bclr	#HF3,x:<<HCR			;inhibit host from reading DSP
	move		#<NOP_OP,n7
	movem		n7,p:<HOST_TX		;NOP the first word in tx vector
	movem		n7,p:<HOST_TX+1		;NOP the second word in tx vector
	movep		n7,x:<<HTX		;in case the HTX interrupt is pending
						;  when this host command is serviced
	move		x0,x:<host_save
	move		#>JSR_OP,x0
	move		#get_tx_rx_ptr,r7
	movem		r7,p:<HOST_RX+1
exit_get_rx_ptr
	movem		x0,p:<HOST_RX
	move		x:<host_save,x0
	rti


get_tx_rx_ptr
	move		x0,x:<host_save
	movep		x:<<HRX,x0		;mem_space is in bits 16 & 17
	move		x0,r7			;moves 16 bit address only
	move		#<NOP_OP,n7
	movem		n7,p:<HOST_RX+1		;NOP the second word in rx vector
	jclr	#23,x0,get_tx_ptr

get_rx_ptr					;DSP rx means host write
	jclr	#17,x0,not_rx_p_space
	move		#>RX_p_OP,x0
	jmp	exit_get_rx_ptr
not_rx_p_space
	jclr	#16,x0,not_rx_y_space
	move		#>RX_y_OP,x0
	jmp	exit_get_rx_ptr
not_rx_y_space
	move		#>RX_x_OP,x0
	jmp	exit_get_rx_ptr

get_tx_ptr					;DSP tx means host read
	movem		n7,p:<HOST_RX		;NOP the DSP rx (host write)
	movep		n7,x:<<HTX		;in case both HTX buffers are empty
	jclr	#17,x0,not_tx_p_space
	movep		p:(r7)+,x:<<HTX		;prefetch and tx first word
	move		#>TX_p_OP,x0
exit_get_tx_ptr
	movem		x0,p:<HOST_TX
	move		x:<host_save,x0
	bset	#HF3,x:<<HCR			;inform host that HTX buffer is
	rti					;  written and it's okay to read.
not_tx_p_space
	jclr	#16,x0,not_tx_y_space
	movep			y:(r7)+,x:<<HTX	;prefetch and tx first word
	move		#>TX_y_OP,x0
	jmp	exit_get_tx_ptr
not_tx_y_space
	movep		x:(r7)+,x:<<HTX		;prefetch and tx first word
	move		#>TX_x_OP,x0
	jmp	exit_get_tx_ptr


set_aux_pot_gain
	move	x0,x:<host_save		; save registers
	move	#>$0084,x0		; set "amps on" bit, select AUX pot
	jmp	cont_set_pot_gain
	
set_input_pot_gain
	move	x0,x:<host_save		; save registers
	move	#>$0082,x0		; set "amps on" bit, select IN pot

cont_set_pot_gain
	movem	a1,p:save_a1
	movep	x:<<HRX,a1		; get the gain value

set_digital_pot
					; a1 = AABBxx  AA- left vol  BB- right vol  (FF-max 00-min)
	movep	x0,y:$FFC0		; enable digital pot
	do	#16,clk_bits_out
	bclr	#4,x0			; initially set data out bit to zero
	rol	a			; rotate next gain bit into place
	jcc	bit_zero		; if carry is zero then data out bit is zero
	bset	#4,x0			; else set data out bit to one
bit_zero
	movep	x0,y:$FFC0		; output data bit with clk low
	bset	#3,x0			; set clk bit
	movep	x0,y:$FFC0		; clock data bit into digital pot
	bclr	#3,x0			; clear clk bit
	movep	x0,y:$FFC0		; reset clk
clk_bits_out
	movep	#$0080,y:$FFC0		; return to original state
	
	movem	p:save_a1,a1		; restore registers
	move	x:<host_save,x0
	rti				; return from interrupt

save_a1	dc	0

;	org	p:

sys_code_image_end


		end

