/*
 * Copyright (c) 1998 IEM
 *
 * MAAB help routines to access the board with/without maab driver
 *
 * purpose: test for MAAB PCI-Card and get memory-space of DSP
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <asm/page.h>

#include "maablib.h"
static int alloc_dsp_memory(void);
static void dealloc_dsp_mem(void);


/* memory for mapping */
pci_t *dsp_mem = NULL;
pci_t *dsp_mem_alloc = NULL;

/* dev file descriptors */
int mem_fd; /* mem */
int dev_fd; /* maab0 */

/* if maab0 or memory */
int mem_not_device = MAAB_NO_DEVICE;

/* diverses */
int maab_verbose = MAAB_VERBOSE;
int maab_debug = MAAB_DEBUG;

/* ******* MAAB init/exit routine (with/without maab device) *********** */
int maab_init(void)
{
  int retval;
  int x;

  mem_not_device = MAAB_DEVICE;
  iopl(0);
  /* first try device */
  if((retval=maab_init_maab()) != MAAB_DEV_OK){
    
    if(maab_debug)
      fprintf(stderr,"maab_init maab open no \n");

    /* ok try  memory map */
    if((retval=maab_init_mem()) != MAAB_MEM_OK)
      return retval;

    mem_not_device = MAAB_NO_DEVICE;
  }

  if(maab_verbose){

    if(mem_not_device == MAAB_NO_DEVICE)
      fprintf(stderr,"MAAB init: try MAAB in mem\n");
    else
      fprintf(stderr,"MAAB init: MAAB device  found\n");


    x=maab_r_hcvr();
    maab_print_hcvr(stderr,x);

    x = maab_r_hstr();
    maab_print_hstr(stderr,x);

    x = maab_r_hctr();
    maab_print_hctr(stderr,x);


    fprintf(stderr,"test write HSTR: HF3 HF4 HF5 with 1 1 1\n");
    x = maab_write(HSTR,HSTR_HF3|HSTR_HF4|HSTR_HF5);
    x = maab_r_hstr();

    fprintf(stderr,"test read  %04x:  %d   %d   %d\n",x,
				is_set(x,HSTR_HF3),
				is_set(x,HSTR_HF4),
				is_set(x,HSTR_HF5)
				);

    /* write/read test */

	 if((x & HCTR_HSM) != 0)
		fprintf(stderr,"Semphores set ? - any other using MAAB\n");


    fprintf(stderr,"test write HCTR: HS0 HS1 HS2 with 1 1 1\n");
    x = maab_w_hctr(maab_r_hctr()|HCTR_HS0|HCTR_HS1|HCTR_HS2);

    x = maab_r_hctr();
    fprintf(stderr,"test read  %04x:  %d   %d   %d\n",x,
				is_set(x,HCTR_HS0),
				is_set(x,HCTR_HS1),
				is_set(x,HCTR_HS2)
				);

    fprintf(stderr,"test write HCTR: HS0 HS1 HS2 with 0 1 0\n");
    x = maab_w_hctr( (x|HCTR_HS1) & ~(HCTR_HS0|HCTR_HS2));
    x = maab_r_hctr();
    fprintf(stderr,"test read  %04x:  %d   %d   %d\n",x,
				is_set(x,HCTR_HS0),
				is_set(x,HCTR_HS1),
				is_set(x,HCTR_HS2)
				);

    fprintf(stderr,"resetting  HCTR: HS0 HS1 HS2 with 0 0 0\n");
    x = maab_r_hctr();
    x = maab_w_hctr( x & ~(HCTR_HS0|HCTR_HS1|HCTR_HS2));

    x = maab_r_hctr();
	 maab_print_hctr(stderr,x);
	 fprintf(stderr,"maabinit: testing ready ....\n\n");


  }
  return MAAB_OK;
}

void maab_exit(void)
{
  if(  mem_not_device == MAAB_DEVICE )
    maab_exit_maab();
  else
    maab_exit_mem();
}


/* ******* MAAB init/exit routine (with maab device) *********** */
int maab_init_maab(void)
{
  int retval;

  if((retval=alloc_dsp_memory()) !=  MAAB_OK)
    return retval;

  dev_fd = -1; /* no sucess */

  if( (dev_fd = open(MAAB_DEVICE_NAME,O_RDWR)) >= 0 ){

	 if(maab_debug) 
		fprintf(stderr,"init_maab: Open %s device\n",
				  MAAB_DEVICE_NAME); 
  
  } 
  else if( (dev_fd = open(MAAB_LOCAL_DEVICE,O_RDWR)) >= 0){

	 if(maab_debug)
		fprintf(stderr,"init_maab: Open device %s \n",MAAB_LOCAL_DEVICE);
  }
  else if(( dev_fd = open(MAAB_ROOT_DEVICE,O_RDWR) ) >= 0 ){

	 if(maab_debug)
		fprintf(stderr,"init_maab: Open %s device\n",MAAB_ROOT_DEVICE);

  }
  else{

	 dealloc_dsp_mem();  
	 fprintf(stderr,"WARNING: cannot open %s device\n",MAAB_DEVICE_NAME);

	 return(MAAB_DEV_OPEN_ERROR);
  };


  if((dsp_mem = (pci_t *) mmap((caddr_t) dsp_mem,
			       MAAB_MEM_SIZE,
			       PROT_WRITE|PROT_READ,MAP_SHARED|MAP_FILE,
			       dev_fd,0)) < 0){

    fprintf(stderr,"ERROR init_maab: MMAP error\n");  
    dealloc_dsp_mem();  
    close(mem_fd);
    return(MAAB_DEV_MAPPING_ERROR);
  }

  if(maab_debug)
    fprintf(stderr,"maab_init_maab mmap  ok\n");

  return MAAB_DEV_OK;
}


void maab_exit_maab(void)
{
  if(dsp_mem != NULL)
    munmap(dsp_mem,MAAB_MEM_SIZE);

  dealloc_dsp_mem();  

  close(dev_fd);
}

/* ******* MAAB mem init/exit routine (without maab device) *********** */
int maab_init_mem()

{
  int retval;

  if(maab_debug)
    fprintf(stderr,"maab_init_mem\n");

  if((retval=alloc_dsp_memory()) !=  MAAB_OK)
    return retval;

  if(maab_debug)
    fprintf(stderr,"maab_init_mem alloc ok\n");


  /* try to open the mem device, if you have permissions */
  if((mem_fd = open("/dev/mem",O_RDWR)) < 0){

    fprintf(stderr,
	    "MAAB init_mem: /dev/mem open: no memory ops allowed\n");

    dealloc_dsp_mem();
    return(MAAB_MEM_OPEN_ERROR);
  }


  if(maab_debug)
    fprintf(stderr,"maab_init_mem open ok 0x%lx\n",(unsigned long) dsp_mem);

  /* now map at BASE address */
  
  dsp_mem=(pci_t *) mmap((caddr_t) dsp_mem,
			 MAAB_MEM_SIZE,
			 PROT_WRITE|PROT_READ,MAP_SHARED|MAP_FIXED,
			 mem_fd, 
			 (off_t) MAAB_MEM_BASE
			 );



  if((long) dsp_mem < 0){

    fprintf(stderr,"MAAB init_mem: dsp mem mapping error\n");

    dealloc_dsp_mem();
    close(mem_fd);
    return(MAAB_MEM_MAPPING_ERROR);

  }

  if(maab_debug)
    fprintf(stderr,"maab_init_mem mmap ok\n");

  return(MAAB_MEM_OK);
}

void maab_exit_mem()
{
  if(dsp_mem != NULL)
    munmap(dsp_mem,MAAB_MEM_SIZE);

  dealloc_dsp_mem();  
  close(mem_fd);
}



/* ----------------- I / O -------------------- */
pci_t maab_read(int addr)
{
  return( dsp_mem[addr] ); 
}

pci_t maab_write(int addr, pci_t x)
{
  return(dsp_mem[addr] = x);
}

/* ---------------- format helps ----------------- */

void maab_print_hstr(FILE *stream,unsigned int x)
{
  fprintf(stream,
     "STATUS   REGISTER: HREQ HINT HF5 HF4 HF3 HRRQ HTRQ TRDY\n");
  fprintf(stream,
     "base(0x14)=0x%04x:  %d    %d    %d   %d   %d   %d    %d    %d\n",
	  x,is_set(x,HSTR_HREQ),
	  is_set(x,HSTR_HINT),
	  is_set(x,HSTR_HF5),
	  is_set(x,HSTR_HF4),
	  is_set(x,HSTR_HF3),
	  is_set(x,HSTR_HRRQ),
	  is_set(x,HSTR_HTRQ),
	  is_set(x,HSTR_TRDY)
	  );
}

void maab_print_hcvr(FILE *stream,unsigned int x)
{
  fprintf(stream,"Command Vector=0x%02x: HC:%d, HMNI:%d\n",
	  (x & HCVR_HV)>>1,
	  is_set(x,HCVR_HC),
	  is_set(x,HCVR_HNMI)
	  );
}



void maab_print_hctr(FILE *stream,unsigned int x)
{
  fprintf(stream,
"COMMAND  REGISTER: base(0x10)=0x%04x:\n" 
"TREQ RREQ HF0 HF1 HF2 DMAE SFT HTF0 HTF1 HRF0 HRF1 HS0 HS1 HS2  TWSD\n",
x);
  fprintf(stream,
"%d    %d    %d   %d   %d    %d   %d   %d    %d    %d    %d    %d   %d   %d    %d\n",
	  is_set(x,HCTR_TREQ),
	  is_set(x,HCTR_RREQ),
	  is_set(x,HCTR_HF0),
	  is_set(x,HCTR_HF1),
	  is_set(x,HCTR_HF2),
	  is_set(x,HCTR_DMAE),
	  is_set(x,HCTR_SFT),
	  is_set(x,HCTR_HTF0),
	  is_set(x,HCTR_HTF1),
	  is_set(x,HCTR_HRF0),
	  is_set(x,HCTR_HRF1),
	  is_set(x,HCTR_HS0),
	  is_set(x,HCTR_HS1),
	  is_set(x,HCTR_HS2),
	  is_set(x,HCTR_TWSD)
	  );
}


void maab_print_hctr_hf012(FILE *stream,unsigned int x)
{
  fprintf(stream,"HCTR: HF0 HF1 HF2  \n");
  fprintf(stream,"%04x:  %d   %d   %d\n",x,
	  is_set(x,HCTR_HF0),
	  is_set(x,HCTR_HF1),
	  is_set(x,HCTR_HF2)
	  );
}

void maab_print_hstr_hf345(FILE *stream,unsigned int x)
{
    fprintf(stream,"HSTR: HF3 HF4 HF5\n");
    fprintf(stream,"%04x:  %d   %d   %d\n",x,
	    is_set(x,HSTR_HF3),
	    is_set(x,HSTR_HF4),
	    is_set(x,HSTR_HF5)
	    );
}

/* ******************** HELPS  ***************** */
static int alloc_dsp_memory(void)
{
  /* allocate a Page for Mapping on a PAGE border */
  /* Idont know why we need this for mapping, 
     but it seems every one does it this way, so I do,...
     by the way it works without on a PC, but on an alpha ?
  */

  if((dsp_mem_alloc=(pci_t *) malloc((MAAB_MEM_SIZE)+(PAGE_SIZE-1)))==NULL){
    fprintf(stderr,"MAAB init_mem: DSP mem allocation error\n");
    return(MAAB_MALLOC_ERROR);
  }
 
  dsp_mem = dsp_mem_alloc;

  if(maab_debug)
    fprintf(stderr,"dsp_alloc %lx, dspmem %lx\n",(unsigned long) dsp_mem_alloc,
	    (unsigned long) (dsp_mem + (MAAB_MEM_SIZE)+(PAGE_SIZE-1)));

  if((unsigned long) dsp_mem %PAGE_SIZE)
    (long) dsp_mem += PAGE_SIZE-((unsigned long)dsp_mem % PAGE_SIZE);

  if(maab_debug)
    fprintf(stderr,"dsp_alloc %lx, dspmem %lx\n",
				(unsigned long) dsp_mem_alloc,(unsigned long) dsp_mem);

  return MAAB_OK;
}

static void dealloc_dsp_mem(void)
{
  if(maab_debug)
    fprintf(stderr,"maab dealloc dspmem %lx\n",(unsigned long) dsp_mem_alloc);

  if(dsp_mem != NULL)
    free(dsp_mem_alloc);

  if(maab_debug)
    fprintf(stderr,"maab dealloc dspmem ok\n");

  dsp_mem = (pci_t *) 0l;
  return;
}


/* ============================ TEST =========================== */
#ifdef MAAB_LIB_TEST

main()
{
printf("Test maab biblio\n");

if(maab_init()==MAAB_OK)
  maab_exit();
else
  printf("somsthing wrong happened\n");
return;
}

#endif
