/*
 * Copyright (c) 1998 IEM
 *
 * maablib.h
 * MAAB Defines for configuration and Memory registers
 *
 * purpose: test for MAAB PCI-Card and get memory-space of DSP
 *
 */

#ifndef MAABLIB_H
#define MAABLIB_H

#include <stdio.h>
/* Specific defs */


#define MAAB_VERBOSE 1
#define MAAB_DEBUG   0

#define MAAB_VendorId 0x1057 /* Motorola */
#define MAAB_DeviceId 0x1801 /* DSP 56301 */

#define MAAB_DEVICE_NAME "maab0"
#define MAAB_LOCAL_DEVICE "../dev/maab0"
#define MAAB_ROOT_DEVICE  "/dev/maab0"


/* predicted BASE ADDRESS ??? */
#define MAAB_MEM_BASE 0xfbff0000l
#define MAAB_MEM_SIZE 0x10000


/* MAAB pci_init routine */
#define MAAB_PCI_OK           0
#define MAAB_PCI_ERROR       -1
#define MAAB_PCI_NO_PCI      -2
#define MAAB_PCI_CANNOT_FIND -3
#define MAAB_PCI_IO_ERROR    -4



/* Typedefs and defines */
#define pci_t  unsigned int

#define MAAB_DEVICE    1
#define MAAB_NO_DEVICE 0

/* MAAB init/exit with/without MAAB device */
int maab_init(void);
void maab_exit(void);

#define MAAB_OK 0

/* MAAB mem init/exit routine (without maab device) */
int maab_init_mem(void);
void maab_exit_mem(void);

#define MAAB_MEM_OK             0
#define MAAB_MALLOC_ERROR      -10
#define MAAB_MEM_OPEN_ERROR    -11
#define MAAB_MEM_MAPPING_ERROR -12

/* MAAB init/exit with MAAB device */
int maab_init_maab(void);
void maab_exit_maab(void);

#define MAAB_DEV_OK               0
#define MAAB_DEV_OPEN_ERROR     -15
#define MAAB_DEV_MAPPING_ERROR  -16

/* global access to memspace */
pci_t maab_read(int addr);
pci_t maab_write(int addr, pci_t x);

/* MAAB REGISTERS */
#define is_set(x,mask) (((x)&(mask))?1:0)

/* MAAB read/write DSP Host-Control REGISTER */
#define HCTR (0x10>>2)

/* Bit 0 reserved */
#define HCTR_TREQ  0x00000002 /* bit 1 */
#define HCTR_RREQ  0x00000004 /* bit 2 */
#define HCTR_HF0   0x00000008 /* bit 3 */
#define HCTR_HF1   0x00000010 /* bit 4 */
#define HCTR_HF2   0x00000020 /* bit 5 */
#define HCTR_DMAE  0x00000040 /* bit 6 */
#define HCTR_SFT   0x00000080 /* bit 7 */
#define HCTR_HTF0  0x00000100 /* bit 8 */
#define HCTR_HTF1  0x00000200 /* bit 9 */
/* Bit 10 reserved */
#define HCTR_HRF0  0x00000800 /* bit 11 */
#define HCTR_HRF1  0x00001000 /* bit 12 */
/* Bit 13 reserved */
#define HCTR_HS0   0x00004000 /* bit 14 */
#define HCTR_HS1   0x00008000 /* bit 15 */
#define HCTR_HS2   0x00010000 /* bit 16 */
#define HCTR_HSM   0x0001C000 /* HS mask */
/* Bit 17,18 reserved */
#define HCTR_TWSD  0x00080000 /* bit 19 */
/* Bit 20-31 reserved */

#define maab_r_hctr()  maab_read(HCTR)
#define maab_w_hctr(x) maab_write(HCTR,(x))

/* MAAB read DSP Host-Status REGISTER */
#define HSTR (0x14>>2)
#define HSTR_TRDY  0x00000001 /* bit 0 */
#define HSTR_HTRQ  0x00000002 /* bit 1 */
#define HSTR_HRRQ  0x00000004 /* bit 2 */
#define HSTR_HF3   0x00000008 /* bit 3 */
#define HSTR_HF4   0x00000010 /* bit 4 */
#define HSTR_HF5   0x00000020 /* bit 5 */
#define HSTR_HINT  0x00000040 /* bit 6 */
#define HSTR_HREQ  0x00000080 /* bit 7 */
/* Bit 8-31 reserved */

#define maab_r_hstr()  maab_read(HSTR)

/* MAAB read/write DSP Host-Command Vector REGISTER */
#define HCVR (0x18>>2)
#define HCVR_HC   0x00000001 /* bit 0 */
#define HCVR_HV   0x000000FE /* bit 1-7 */
/* Bit 8-14 reserved */
#define HCVR_HNMI 0x00008000 /* bit 15 */

#define maab_r_hcvr()  maab_read(HCVR)
#define maab_w_hcvr(x) maab_write(HCVR,(x))

#define maab_r_hcvector()   ((maab_r_hcvr() & HCVR_HV)>>1)
#define maab_w_hcvector(v)  (maab_w_hcvr( ((v)<<1) & HCVR_HV))

/* MAAB read/write DSP DATA REGISTER */
#define HTXR (0x1C>>2)
#define HRXS (0x1C>>2)


#define maab_r_data()  ((int) maab_read(HRXS))
#define maab_w_data(x) ((int) maab_write(HTXR,(x)))

/* write to mem map data */
#define maab_r_mdata(a)  ((int) maab_read(HRXR+a))
#define maab_w_mdata(a,x) ((int) maab_write(HTXR+a),(x)))


/* ---------------- format helps ----------------- */

void maab_print_hstr(FILE *stream,unsigned int x);
void maab_print_hcvr(FILE *stream,unsigned int x);
void maab_print_hctr(FILE *stream,unsigned int x);
void maab_print_hctr_hf012(FILE *stream,unsigned int x);
void maab_print_hstr_hf345(FILE *stream,unsigned int x);
#endif
